from cent import Client, ClientNotEmpty, RequestException
import csv
from lcutils import flexi_csv_reader
import logging
import os
import signal
import time
from typing import Union
from watchdog.events import FileSystemEventHandler, FileSystemEvent
from ..persistence import PersistLogs
from ..pygtail import Pygtail
from ..settings import BUFFER_LINES, CLOUD_SERVER, FILTER_SERVER
# use a larger threshold if cloud server, as the activity will be much higher
BUFFER_SIZE = BUFFER_LINES if CLOUD_SERVER else BUFFER_LINES / 2

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------
class LogTailerHandler(FileSystemEventHandler):
    """
    Base Log Tailer handler.

    When tailing a log file to save lines to database, we need
    to manage the last position so we can start there at the
    next launch time.
    """

    def __init__(self, ws_client):
        self._last_position: int = 0
        self._file_handle = None
        self.ws_client: Client = ws_client
        self.persist_client: PersistLogs = self.persist_class() if self.persist_class else None

        # on startup, assume no ws presence until _last_presence_check expires
        self._websocket_has_presence: bool = False
        self._last_presence_check: float = time.time()

        self.post_to_console_api = bool(self.persist_class)

        self._tailer = None

        # Buffer lines before POSTing to API handler
        self._api_buffer = []

        self._watched_file = None

    def on_closed(self, event: FileSystemEvent):
        self.terminate_handler(event)

    def on_deleted(self, event: FileSystemEvent):
        self.terminate_handler(event)

    def on_moved(self, event: FileSystemEvent):
        self.terminate_handler(event)

    def on_modified(self, event: FileSystemEvent):
        """
        Read all new lines on receiving the "modified" event.
        """

        # sleep a bit to allow more lines to accumulate
        time.sleep(.3)

        for line in self.tailer:
            lld = self.as_data(line)
            if not lld:
                continue
            self.publish_to_websocket(lld, line)
            self.persist_data(lld, line)

    def terminate_handler(self, event: FileSystemEvent):
        """
        Catch deletes. The task manager should restart to watch the new inode.
        """
        logger.info('Terminating - received event: %s' % event)
        os.kill(os.getgid(), signal.SIGTERM)

    def parsed_ws_message(self, ll: dict, line: str):
        """
        Override on subclasses to customize parsing and the lines included.
        """
        return ll

    def websocket_has_presence(self) -> bool:
        """
        If websocket server has no user present, then there's
        no reason to publish messages to the server.

        Check for presence only every 5 seconds as modified
        events will occur much oftener than that.
        """
        # if no channel is specified, then we're not publishing messages
        if not self.ws_channel:
            return False

        if time.time() - self._last_presence_check > 5:
            try:
                self._websocket_has_presence = self.ws_client.presence(self.ws_channel)
            except ClientNotEmpty:
                self._websocket_has_presence = True
            except RequestException:
                self._websocket_has_presence = False

        return self._websocket_has_presence

    def publish_to_websocket(self, ll: dict, line: str):
        """
        Publish log lines to websocket server.
        """
        if not self.websocket_has_presence():
            return

        ll = self.parsed_ws_message(ll, line)
        if ll:
            try:
                self.ws_client.publish(self.ws_channel, ll)
            except Exception as e:
                logger.exception(f'Unable to publish {line!r} to websocket: {e}')

    def parsed_persistable_data(self, ll: dict, line: str):
        """
        Override on subclasses to customize parsing and the lines included.
        """
        return ll

    def persist_data(self, ll: dict, line: str):
        """
        Accumulate log lines and then post console
        task scheduler to save them to the database.
        """
        if not FILTER_SERVER or not self.persist_client:
            return

        parsed = self.parsed_persistable_data(ll, line)
        if not parsed:
            return

        self._api_buffer.append(parsed)

        if len(self._api_buffer) > BUFFER_SIZE:

            resp = self.persist_client.save(self._api_buffer)
            self._api_buffer = []

            # if we're not getting good responses from the console
            # don't save file position, so that we reread the file
            # presumably after fixing the error
            if resp is not None:
                return

            self.tailer._update_offset_file()

    def get_watch_file(self):
        """
        Override on subclasses in case any special logic is required.
        Basically for redwood's errors.log file which doesn't close
        when log-rotated, so we need to find the opened file.
        """
        return self.watch_file

    def as_data(self, line: str) -> Union[dict, str]:
        """
        Parse a line and return it as dictionary.
        Override on subclass mixins to customize.
        """
        if not self.csv_fields or not line:
            return line
        try:
            return flexi_csv_reader(csv.DictReader((line,), fieldnames=self.csv_fields)).__next__()
        except (RuntimeError, StopIteration) as e:
            logger.info(f'Malformed logline: {line}')
            logger.exception(e)
            return ''

    @property
    def tailer(self) -> Pygtail:
        """
        Open file object that reads new log lines.
        """
        track_position = bool(self.offset_file)
        if not self._tailer:
            self._tailer = Pygtail(
                self.get_watch_file(),
                every_n=BUFFER_LINES,
                offset_file=self.offset_file or None,
                read_from_end=not track_position,
                manual_offset_update=track_position,
                full_lines=True,
            )
        return self._tailer


__all__ = [
    'LogTailerHandler',
]
