from ..parsers import parse_access_log, scrub_logline, skip_malformed
from ..persistence import PersistRedwoodLogs
from ..settings import REDWOOD_LOG_FIELDS, TLS_LOG_FIELDS


# ----------------------------------------------------------------------
class RedwoodAccessMixin:
    """
    Tail Redwood Access log
    """
    name = 'redwood'
    ws_channel = 'filter_access'
    offset_file = '/var/tmp/redwood.offset'
    watch_file = '/var/log/redwood/access.log'
    persist_class = PersistRedwoodLogs
    csv_fields = REDWOOD_LOG_FIELDS

    def parsed_persistable_data(self, ll: dict, line: str):
        """
        Return log line data if it's well formed and passes "scrub" test.
        """
        if not scrub_logline(ll) and not skip_malformed(ll):
            return ll

    def parsed_ws_message(self, ll: dict, line: str) -> dict:
        """
        Scrub out all lines that shouldn't be sent to websocket server.
        Don't call scrub_logline here, as we need to send ssl-bump
        lines to the websocket server.
        """
        if not skip_malformed(ll):
            return parse_access_log(ll)


# ----------------------------------------------------------------------
class RedwoodErrorsMixin:
    """
    Tail Redwood Errors log
    """
    name = 'errors'
    ws_channel = 'filter_errors'
    offset_file = '/var/tmp/redwood_errors.offset'
    watch_file = '/var/log/redwood/errors.log'
    persist_class = None
    csv_fields = ()

    def parsed_ws_message(self, ll: dict, line: str) -> dict:
        """
        Send errors.log to websocket server,
        with no further line parsing.
        """
        return {'line': line}


# ----------------------------------------------------------------------
class RedwoodTLSMixin:
    """
    Tail Redwood TLS Log
    """
    name = 'tls'
    ws_channel = 'filter_tls'
    offset_file = '/var/tmp/redwood_tls.offset'
    watch_file = '/var/log/redwood/tls.log'
    persist_class = None
    csv_fields = TLS_LOG_FIELDS

    def parsed_ws_message(self, ll: dict, line: str) -> dict:
        return {
            'server': ll['serverAddr'],
            'userip': ll['user'],
            'line': line,
        }


__all__ = (
    'RedwoodAccessMixin',
    'RedwoodErrorsMixin',
    'RedwoodTLSMixin',
)
