import csv
from lcutils import flexi_csv_reader
from lchttp.uri import decode
import logging
from logtailer.settings import (
    LC_CAT_PREFIX,
    NON_SCORERS,
)

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------
def flexi_csv_reader(csv_reader):
    """
    Skip to next line when encountering a CSV error
    instead of blowing up. Log Lines aren't that
    important; keep moving.
    """
    lines = []

    for line in csv_reader:
        try:
            lines.append(line)
        except csv.Error:
            logger.error(f'Malformed line: {line}')
            continue

    return lines


# ---------------------------------------------------------------
def acl_dict(line: str) -> dict[str, int]:
    """
    Convert tally / score string to dictionary, where string is

    'lc_auto 23543, lc_racing 423, racing 233, sports 158'

    '< guitar> 815, <electric guitar> 244, <name search> 122, <logistics> 60'

    >>> acl_dict('lc_auto 23543, lc_racing 423, racing 233, sports 158')
    {'< guitar>': 815, '<electric guitar>': 244, '<name search>': 122, '<logistics>': 60}

    """
    if not line:
        return {}

    acldict = {}

    for acl in line.split(','):
        acl = acl.strip()
        try:
            rspace = acl.rfind(' ')
            name = acl[:rspace]
            acldict[name] = int(acl[rspace:])
        except (TypeError, ValueError):
            logger.error(f'{line} is malformed. Unable to parse text & score.')

    return acldict


# ----------------------------------------------------------------------
def log_line_dicts(logs: list, fields) -> list[dict]:
    """Return list of dictionaries from log lines"""

    if isinstance(logs, str):
        logs = logs.splitlines()

    return flexi_csv_reader(csv.DictReader(logs, fields))


# ----------------------------------------------------------------------
def log_line_dict(log: str, fields) -> dict:
    """Return dictionary from log line"""

    return flexi_csv_reader(csv.reader([log], fields))[0]


# ----------------------------------------------------------------------
def top_score(conditions: list[str], scores: dict[str, int]) -> str:
    """
    Extract Top-scoring category & score, by looking at conditions
    and getting highest interesting category there. If no interesting
    category found, then loop through scores and get first category
    of interest.
    """
    for acl in conditions:
        if acl in NON_SCORERS:
            continue

        try:
            return f'{acl} {scores[acl]}'
        except KeyError:
            break

    for cat, score in scores.items():
        if cat not in NON_SCORERS:
            return f'{cat} {score}'

    return ''


# ----------------------------------------------------------------------
def parse_access_log(ll: dict) -> dict:
    """
    Parse an access.log line before sending to WS server

    :param ll: dict from CSV-parsed Redwood access.log line
    :return: Dict of data ready to publish on WS server
    """
    conditions = [c.strip() for c in ll['conditions'].split()]

    # If no page title, copy dict to add placeholder values for
    # the WS server - but copy so that we don't pollute data for
    # persisting to database.
    if not ll['page_title']:
        ll = ll.copy()

    ll.update({
        'date': ll['ldate'].split()[1],
        'top_score': top_score(conditions, acl_dict(ll['scores'])),
        'ua': decode(ll['user_agent'].strip()) or 'No UA Provided',
        'page_title': ll['page_title'] or 'No Page Title',
    })

    # if all categories aren't custom or line not blocked, hide rule Tally
    if not ll['conditions'].startswith(LC_CAT_PREFIX) or ll['action'] != 'block':
        ll['tally'] = ''

    log = '{date} {user} {action} {url} {method} {content_type} (ACLs: {conditions}) "{ua}" "{page_title}" [{tally}] ({top_score})'

    try:
        rule_description = ll['rule_description']
        record_prefix, pk = rule_description.split('/')
    except (AttributeError, KeyError, IndexError, ValueError):
        record_prefix = pk = ''

    return {
        'action': ll['action'],
        'userip': ll['user'],
        'line': log.format(**ll),
        'pk': pk,
        'record': record_prefix,
    }


__all__ = (
    'acl_dict',
    'log_line_dicts',
    'log_line_dict',
    'parse_access_log',
    'top_score',
)
